open ERTL

(* These functions allow turning an [ERTL] control flow graph into an
   explicit graph, that is, making successor edges explicit. This is
   useful in itself and facilitates the computation of predecessor
   edges. *)

let instruction_successors (i : instruction) =
  match i with
  | IReturn _ ->
      Label.Set.empty
  | INewFrame l
  | IDeleteFrame l
  | IGetHwReg (_, _, l)
  | ISetHwReg (_, _, l)
  | IGetStack (_, _, l)
  | ISetStack (_, _, l)
  | IGetGlobal (_, _, l)
  | ISetGlobal (_, _, l)
  | IConst (_, _, l)
  | IUnOp (_, _, _, l)
  | IBinOp (_, _, _, _, l)
  | ICall (_, _, l)
  | ILoad (_, _, _, l)
  | IStore (_, _, _, l)
  | IGoto l ->
      Label.Set.singleton l
  | IUnBranch (_, _, l1, l2)
  | IBinBranch (_, _, _, l1, l2) ->
      Label.Set.couple l1 l2

let successors (graph : graph) : Label.SetMap.t =
  Label.Map.map instruction_successors graph

(* The analysis uses the lattice of pairs of a subset of the
   function's pseudo-registers and a subset of the hardware
   registers. The lattice's join operation is pointwise set union,
   which reflects the fact that a pseudo-register or hardware register
   is deemed live at a program point if and only if it is live at any
   of the successors of that program point. *)

module L = struct

  type t =
      Register.Set.t * MIPS.RegisterSet.t

  let bottom =
    Register.Set.empty, MIPS.RegisterSet.empty

  let psingleton r =
    Register.Set.singleton r, MIPS.RegisterSet.empty

  let hsingleton hwr =
    Register.Set.empty, MIPS.RegisterSet.singleton hwr

  let join (rset1, hwrset1) (rset2, hwrset2) =
    (Register.Set.union rset1 rset2, MIPS.RegisterSet.union hwrset1 hwrset2)

  let diff (rset1, hwrset1) (rset2, hwrset2) =
    (Register.Set.diff rset1 rset2, MIPS.RegisterSet.diff hwrset1 hwrset2)

  let equal (rset1, hwrset1) (rset2, hwrset2) =
    Register.Set.equal rset1 rset2 && MIPS.RegisterSet.equal hwrset1 hwrset2

end

(* These are the sets of pseudo-registers and hardware registers
   defined at (written by) an instruction. *)

let defined (i : instruction) : L.t = 
  failwith "Pas implment !"

(* This is the set of pseudo-registers and hardware registers used at
   (read by) an instruction. *)

let used (i : instruction) : L.t = 
  failwith "Pas implment !"

(* An instruction is considered pure if it has no side effect, that
   is, if its only effect is to write a value to its destination
   pseudo-register.

   [eliminable liveafter i] returns [Some l], where [l] is [i]'s single
   successor, if instruction [i] is eliminable. Otherwise, it returns
   [None]. The parameter [liveafter] is the set of variables that are live
   after the instruction. *)

let eliminable ((pliveafter, hliveafter) : L.t) (i : instruction) =
  failwith "Pas implment !"

(* This is the abstract semantics of instructions. It defines the
   [out] value of an instruction in terms of its [in] value. That is,
   since liveness analysis is a backward analysis, it defines the
   pseudo-registers that are live before the instruction in terms of
   those that are live after the instruction. *)

let instruction_semantics (i : instruction) (invalue : L.t) : L.t =
  failwith "Pas implment !"

(* Let us specialize the data flow analysis problem solver for this
   lattice. *)

module K =
  Kildall.Make (L)

type valuation =
    K.valuation

(* This is how we turn an [ERTL] procedure into a liveness analysis
   problem and solve it. *)

let analyze (proc : procedure) : valuation * valuation =

  (* Turn the procedure's control flow graph into explicit successor
     and predecessor graphs. *)

  let successors = successors proc.graph in
  let predecessors = Label.SetMap.reverse successors in

  (* Formulate and solve the analysis problem. *)

  let module S = K.Simplified (struct

    (* Since this is a backward analysis, an instruction's successors
       in the analysis problem are its predecessors in the control
       flow graph, and vice versa. *)

    let successors label =
      Label.SetMap.find label predecessors

    and predecessors label =
      Label.SetMap.find label successors

    let semantics label =
      instruction_semantics (Label.Map.find label proc.graph)

    (* The seed is bottom everywhere. *)

    let seed label =
      L.bottom

    (* The problem domain is the entire control flow graph. We do not
       assume that all points are reachable from the exit point --
       that would not be true in a procedure with an infinite loop,
       for instance. *)

    let entry =
      Label.Map.domain proc.graph

  end) in

  (* Transmit the solutions. Because this is a backward analysis, the
     [in] solution reported by Kildall's algorithm is our [after]
     solution, and the [out] solution reported by Kildall's algorithm
     is our [before] solution. *)

  S.outsolution, S.insolution
